@@ -8,11 +8,10 @@
 #include <libart_lgpl/art_uta.h>
 
 void
-art_rgba_svp_alpha (const ArtSVP *svp,
-                  int x0, int y0, int x1, int y1,
-                  art_u32 rgba,
-                  art_u8 *buf, int rowstride,
-                  ArtAlphaGamma *alphagamma);
-
-#endif
+gnome_print_art_rgba_svp_alpha (const ArtSVP *svp,
+				int x0, int y0, int x1, int y1,
+				art_u32 rgba,
+				art_u8 *buf, int rowstride,
+				ArtAlphaGamma *alphagamma);
 
+#endif /* SP_ART_RGBA_SVP_H */
