@@ -99,20 +99,19 @@ static int meta_fill (GnomePrintContext *pc, const ArtBpath *bpath, ArtWindRule
 static int meta_stroke (GnomePrintContext *pc, const ArtBpath *bpath);
 static int meta_image (GnomePrintContext *pc, const gdouble *affine, const guchar *px, gint w, gint h, gint rowstride, gint ch);
 static int meta_glyphlist (GnomePrintContext *pc, const gdouble *affine, GnomeGlyphList *gl);
-static int meta_close (GnomePrintContext *pc);
 
 static void meta_color (GnomePrintContext *pc);
 static void meta_line (GnomePrintContext *pc);
 static void meta_dash (GnomePrintContext *pc);
 
-static void gpm_encode_string (GnomePrintContext *pc, const guchar *str);
+static void gpm_encode_string (GnomePrintContext *pc, const gchar *str);
 static void gpm_encode_int (GnomePrintContext *pc, gint32 value);
 static void gpm_encode_double (GnomePrintContext *pc, double d);
-static void gpm_encode_block (GnomePrintContext *pc, const guchar *data, gint size);
+static void gpm_encode_block (GnomePrintContext *pc, const gchar *data, gint size);
 
 static gboolean gpm_ensure_space (GnomePrintMeta *meta, int size);
 static void gpm_encode_bpath (GnomePrintContext *pc, const ArtBpath *bpath);
-static const char *gpm_decode_bpath (const char *data, ArtBpath **bpath);
+static const guchar *gpm_decode_bpath (const guchar *data, ArtBpath **bpath);
 
 static const guchar *gpm_decode_string (const guchar *data, guchar **dest);
 
@@ -164,22 +163,12 @@ gnome_print_meta_class_init (GnomePrintMetaClass *klass)
 	pc_class->stroke = meta_stroke;
 	pc_class->image = meta_image;
 	pc_class->glyphlist = meta_glyphlist;
-
-	pc_class->close = meta_close;
 }
 
 static void
 gnome_print_meta_init (GnomePrintMeta *meta)
 {
-	meta->buf = g_new (guchar, BLOCKSIZE);
-	meta->b_length = 0;
-	meta->b_size = BLOCKSIZE;
-
-	gpm_encode_block (GNOME_PRINT_CONTEXT (meta), METAFILE_SIGNATURE, METAFILE_SIGNATURE_SIZE);
-	gpm_encode_int_header (GNOME_PRINT_CONTEXT (meta), 0);
-
-	meta->page = 0;
-	meta->pagenum = -1;
+	gnome_print_meta_reset (meta);
 }
 
 static void
@@ -194,6 +183,26 @@ meta_finalize (GObject *object)
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+gint
+gnome_print_meta_get_length (const GnomePrintMeta *meta)
+{
+	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), 0);
+
+	return meta->b_length;
+}
+
+static void
+gnome_print_meta_set_length (GnomePrintMeta *meta, guint len)
+{
+	gint32 l;
+
+	g_return_if_fail (GNOME_IS_PRINT_META (meta));
+
+	meta->b_length = len;
+	l = g_htonl (meta->b_length);
+	memcpy (meta->buf + METAFILE_SIGNATURE_SIZE, &l, 4);
+}
+
 static gint
 meta_beginpage (GnomePrintContext *ctx, const guchar *name)
 {
@@ -202,7 +211,7 @@ meta_beginpage (GnomePrintContext *ctx, const guchar *name)
 	meta = GNOME_PRINT_META (ctx);
 
 	/* Save page header position */
-	meta->page = meta->b_length;
+	meta->page = gnome_print_meta_get_length (meta);
 
 	/* Encode page header */
 	gpm_encode_block (ctx, PAGE_SIGNATURE, PAGE_SIGNATURE_SIZE);
@@ -213,9 +222,7 @@ meta_beginpage (GnomePrintContext *ctx, const guchar *name)
 
 	/* Encode beginpage */
 	gpm_encode_int (ctx, GNOME_META_BEGINPAGE);
-	if (!name)
-		name = "";
-	gpm_encode_string (ctx, name);
+	gpm_encode_string (ctx, name ? (const gchar *) name : "");
 
 	return GNOME_PRINT_OK;
 }
@@ -320,7 +327,7 @@ meta_image (GnomePrintContext *pc, const gdouble *affine, const guchar *px, gint
 	}
 
 	for (y = 0; y < h; y++){
-		gpm_encode_block (pc, px, w * ch);
+		gpm_encode_block (pc, (const gchar *) px, w * ch);
 		px += rowstride;
 	}
 
@@ -359,7 +366,7 @@ meta_glyphlist (GnomePrintContext *pc, const gdouble *affine, GnomeGlyphList *gl
 			break;
 		case GGL_FONT:
 			gpm_encode_double (pc, gnome_font_get_size (gl->rules[i].value.font));
-			gpm_encode_string (pc, gnome_font_get_name (gl->rules[i].value.font));
+			gpm_encode_string (pc, (const gchar *) gnome_font_get_name (gl->rules[i].value.font));
 			break;
 		default:
 			break;
@@ -369,21 +376,6 @@ meta_glyphlist (GnomePrintContext *pc, const gdouble *affine, GnomeGlyphList *gl
 	return GNOME_PRINT_OK;
 }
 
-static gint
-meta_close (GnomePrintContext *ctx)
-{
-	GnomePrintMeta *meta;
-	gint32 len;
-
-	meta = GNOME_PRINT_META (ctx);
-
-	len = g_htonl (meta->b_length);
-
-	memcpy (meta->buf + METAFILE_SIGNATURE_SIZE, &len, 4);
-
-	return GNOME_PRINT_OK;
-}
-
 /* Private GPGC methods */
 
 static void
@@ -456,6 +448,24 @@ gnome_print_meta_new (void)
 	return (GnomePrintContext *) meta;
 }
 
+void
+gnome_print_meta_reset (GnomePrintMeta *meta)
+{
+	g_return_if_fail (GNOME_IS_PRINT_META (meta));
+
+	if (meta->buf) g_free (meta->buf);
+	meta->buf = g_new (guchar, BLOCKSIZE);
+	meta->b_length = 0;
+	meta->b_size = BLOCKSIZE;
+
+	gpm_encode_block (GNOME_PRINT_CONTEXT (meta), METAFILE_SIGNATURE,
+			  METAFILE_SIGNATURE_SIZE);
+	gpm_encode_int_header (GNOME_PRINT_CONTEXT (meta), 0);
+
+	meta->page = 0;
+	meta->pagenum = -1;
+}
+
 const guchar *
 gnome_print_meta_get_buffer (const GnomePrintMeta *meta)
 {
@@ -465,15 +475,6 @@ gnome_print_meta_get_buffer (const GnomePrintMeta *meta)
 	return meta->buf;
 }
 
-gint
-gnome_print_meta_get_length (const GnomePrintMeta *meta)
-{
-	g_return_val_if_fail (meta != NULL, 0);
-	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), 0);
-
-	return meta->b_length;
-}
-
 int
 gnome_print_meta_get_pages (const GnomePrintMeta *meta)
 {
@@ -483,14 +484,17 @@ gnome_print_meta_get_pages (const GnomePrintMeta *meta)
 	return meta->pagenum + 1;
 }
 
-static const char *
-decode_int (const char *data, gint32 *dest)
+static const guchar *
+decode_int (const guchar *data, gint32 *dest)
 {
 	guint32 vabs = 0, mask = 0x3f;
 	int bits = 6, shift = 0;
 	int neg;
 	char c;
 
+	g_return_val_if_fail (data, NULL);
+	g_return_val_if_fail (dest, NULL);
+
 	neg = (*data & 0x40);
 	do {
 		vabs |= ((c = *data++) & mask) << shift;
@@ -503,8 +507,8 @@ decode_int (const char *data, gint32 *dest)
 	return data;
 }
 
-static const char *
-decode_double (const char *data, double *dest)
+static const guchar *
+decode_double (const guchar *data, double *dest)
 {
 	int i;
 	data = decode_int (data, &i);
@@ -556,7 +560,6 @@ gpm_render (GnomePrintContext *dest, const guchar *data, gint pos, gint len, gbo
 
 	data = data + pos;
 	end = data + len;
-
 	while (data < end){
 		gint32 opcode, i;
 		guchar *cval;
@@ -731,6 +734,106 @@ gpm_render (GnomePrintContext *dest, const guchar *data, gint pos, gint len, gbo
 	return GNOME_PRINT_OK;
 }
 
+static void
+search_page (const guchar *b, guint b_len, guint page, guint *pos, guint *len)
+{
+	gint pagenum;
+
+	g_return_if_fail (b != NULL);
+	g_return_if_fail (pos != NULL);
+
+	*pos = METAFILE_HEADER_SIZE;
+
+        /* Search the page */
+	for (pagenum = 0; *pos < b_len; pagenum++) {
+		gint32 l;
+
+		if (strncmp ((const gchar *) (b + *pos), PAGE_SIGNATURE, PAGE_SIGNATURE_SIZE))
+			break;
+		gpm_decode_int_header (b + *pos + PAGE_SIGNATURE_SIZE, &l);
+		*pos += PAGE_HEADER_SIZE;
+		if (pagenum == page) {
+			if (len) *len = MIN (b_len - *pos, l);
+			return;
+		}
+		*pos += l;
+	}
+
+	/* We have not found the page. */
+	*pos = b_len;
+	if (len) *len = 0;
+}
+
+/**
+ * gnome_print_meta_render:
+ * @meta:
+ * @dest:
+ *
+ * Render to a specified output context
+ *
+ * Return Value:
+ **/
+gint
+gnome_print_meta_render (const GnomePrintMeta *meta, GnomePrintContext *dst)
+{
+	g_return_val_if_fail (GNOME_IS_PRINT_META (meta),
+			      GNOME_PRINT_ERROR_BADCONTEXT);
+
+	return gnome_print_meta_render_data (dst, meta->buf, meta->b_length);
+}
+
+/**
+ * gnome_print_meta_get_page_name:
+ * @meta:
+ * @page:
+ * @page_name:
+ *
+ * Get the name of a specific page.
+ *
+ * Return Value:
+ **/
+gint
+gnome_print_meta_get_page_name (const GnomePrintMeta *meta, guint page, guchar **page_name)
+{
+	guint pos;
+	gint32 opcode;
+	const guchar *data;
+
+	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), GNOME_PRINT_ERROR_BADCONTEXT);
+	g_return_val_if_fail (page_name != NULL, GNOME_PRINT_ERROR_BADCONTEXT);
+
+	*page_name = NULL;
+
+	/* Search the page */
+	search_page (meta->buf, meta->b_length, page, &pos, NULL);
+	if (pos >= meta->b_length) return GNOME_PRINT_ERROR_BADVALUE;
+
+	data = decode_int (meta->buf + pos, &opcode);
+	if (opcode != GNOME_META_BEGINPAGE) return GNOME_PRINT_ERROR_BADCONTEXT;
+	gpm_decode_string (data, page_name);
+
+	return GNOME_PRINT_OK;
+}
+
+/**
+ * gnome_print_meta_render_page:
+ * @meta:
+ * @dest:
+ * @page:
+ *
+ * Render page to specified output context
+ * 
+ * Return Value:
+ **/
+gint
+gnome_print_meta_render_page (const GnomePrintMeta *meta, GnomePrintContext *dst, gint page, gboolean pageops)
+{
+	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), GNOME_PRINT_ERROR_BADCONTEXT);
+
+	return gnome_print_meta_render_data_page (dst, meta->buf,
+					meta->b_length, page, pageops);
+}
+
 /**
  * gnome_print_meta_render_data:
  * @ctx: 
@@ -750,13 +853,13 @@ gnome_print_meta_render_data (GnomePrintContext *ctx, const guchar *data, gint l
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (ctx), GNOME_PRINT_ERROR_BADCONTEXT);
 	g_return_val_if_fail (data != NULL, GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (length >= METAFILE_HEADER_SIZE, GNOME_PRINT_ERROR_UNKNOWN);
-	g_return_val_if_fail (!strncmp (data, METAFILE_SIGNATURE, METAFILE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
+	g_return_val_if_fail (!strncmp ((const gchar *) data, METAFILE_SIGNATURE, METAFILE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
 
 	pos = METAFILE_HEADER_SIZE;
 
 	while (pos < length) {
 		gint32 len;
-		g_return_val_if_fail (!strncmp (data + pos, PAGE_SIGNATURE, PAGE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
+		g_return_val_if_fail (!strncmp ((const gchar *) (data + pos), PAGE_SIGNATURE, PAGE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
 		gpm_decode_int_header (data + pos + PAGE_SIGNATURE_SIZE, &len);
 		pos += PAGE_HEADER_SIZE;
 		if (len == 0)
@@ -783,32 +886,17 @@ gnome_print_meta_render_data (GnomePrintContext *ctx, const guchar *data, gint l
 gint
 gnome_print_meta_render_data_page (GnomePrintContext *ctx, const guchar *data, gint length, gint page, gboolean pageops)
 {
-	gint pos, pagenum;
+	guint len, pos;
 
 	g_return_val_if_fail (ctx != NULL, GNOME_PRINT_ERROR_BADCONTEXT);
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (ctx), GNOME_PRINT_ERROR_BADCONTEXT);
 	g_return_val_if_fail (data != NULL, GNOME_PRINT_ERROR_UNKNOWN);
 	g_return_val_if_fail (length >= METAFILE_HEADER_SIZE, GNOME_PRINT_ERROR_UNKNOWN);
-	g_return_val_if_fail (!strncmp (data, METAFILE_SIGNATURE, METAFILE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
+	g_return_val_if_fail (!strncmp ((const gchar *) data, METAFILE_SIGNATURE, METAFILE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
 
-	pos = METAFILE_HEADER_SIZE;
-
-	pagenum = 0;
-	while (pos < length) {
-		gint32 len;
-		g_return_val_if_fail (!strncmp (data + pos, PAGE_SIGNATURE, PAGE_SIGNATURE_SIZE), GNOME_PRINT_ERROR_UNKNOWN);
-		gpm_decode_int_header (data + pos + PAGE_SIGNATURE_SIZE, &len);
-		pos += PAGE_HEADER_SIZE;
-		if (len == 0)
-			len = length - pos;
-		if (pagenum == page) {
-			return gpm_render (ctx, data, pos, len, pageops);
-		}
-		pos += len;
-		pagenum += 1;
-	}
-
-	return GNOME_PRINT_ERROR_UNKNOWN;
+	search_page (data, length, page, &pos, &len);
+	if (!len) return GNOME_PRINT_ERROR_BADVALUE;
+	return gpm_render (ctx, data, pos, len, pageops);
 }
 
 /**
@@ -874,7 +962,7 @@ gnome_print_meta_render_file_page (GnomePrintContext *ctx, const guchar *filenam
 }
 
 static void
-gpm_encode_string (GnomePrintContext *pc, const guchar *str)
+gpm_encode_string (GnomePrintContext *pc, const gchar *str)
 {
 	gint bytes;
 
@@ -925,7 +1013,7 @@ gpm_encode_int (GnomePrintContext *pc, gint32 value)
 
 	out[-1] |= 0x80;
 	if (value < 0) out0[0] |= 0x40;
-	meta->b_length = out - meta->buf;
+	gnome_print_meta_set_length (meta, out - meta->buf);
 }
 
 static void
@@ -954,7 +1042,7 @@ gpm_encode_double (GnomePrintContext *pc, double d)
 			}
 #elif G_BYTE_ORDER == G_LITTLE_ENDIAN
 			g_assert (sizeof (double) == 8);
-			gpm_encode_block (pc, (const guchar *) &d, sizeof (d));
+			gpm_encode_block (pc, (const gchar *) &d, sizeof (d));
 #else
 #error encode_double_needs_attention
 #endif
@@ -963,11 +1051,9 @@ gpm_encode_double (GnomePrintContext *pc, double d)
 }
 
 static void
-gpm_encode_block (GnomePrintContext *pc, const guchar *data, gint size)
+gpm_encode_block (GnomePrintContext *pc, const gchar *data, gint size)
 {
-	GnomePrintMeta *meta;
-
-	meta = (GnomePrintMeta *) pc;
+	GnomePrintMeta *meta = GNOME_PRINT_META (pc);
 
 	if (!GPM_ENSURE_SPACE (meta, size)) {
 		g_warning ("file %s: line %d: Cannot grow metafile buffer (%d bytes)", __FILE__, __LINE__, size);
@@ -975,7 +1061,7 @@ gpm_encode_block (GnomePrintContext *pc, const guchar *data, gint size)
 	}
 
 	memcpy (meta->buf + meta->b_length, data, size);
-	meta->b_length += size;
+	gnome_print_meta_set_length (meta, meta->b_length + size);
 }
 
 static gboolean
@@ -1027,13 +1113,17 @@ gpm_encode_bpath (GnomePrintContext *pc, const ArtBpath *bpath)
 	gpm_encode_int (pc, ART_END);
 }
 
-static const char *
-gpm_decode_bpath (const char *data, ArtBpath **bpath)
+static const guchar *
+gpm_decode_bpath (const guchar *data, ArtBpath **bpath)
 {
 	ArtBpath *p;
 	gint32 len, code;
 
 	data = decode_int (data, &len);
+	if (!len) {
+		g_warning ("Could not decode bpath: Corrupt data!");
+		return NULL;
+	}
 	*bpath = g_new (ArtBpath, len);
 
 	p = *bpath;
@@ -1076,7 +1166,7 @@ gpm_encode_int_header (GnomePrintContext *pc, gint32 value)
 
 	new = g_htonl (value);
 
-	gpm_encode_block (pc, (const guchar *) &new, sizeof (gint32));
+	gpm_encode_block (pc, (const gchar *) &new, sizeof (gint32));
 }
 
 static const guchar *
@@ -1089,4 +1179,3 @@ gpm_decode_int_header (const guchar *data, gint32 *dest)
 
 	return data + sizeof (gint32);
 }
-
