@@ -137,10 +137,8 @@ gpa_reference_set_value (GPANode *node, const guchar *value)
 	reference = GPA_REFERENCE (node);
 	old = reference->ref;
 
-	if (G_OBJECT_TYPE (old->parent) != GPA_TYPE_LIST) {
-		g_warning ("Can only set GPAReference values of GPALists");
-		return FALSE;
-	}
+	g_return_val_if_fail (old->parent != NULL, FALSE);
+	g_return_val_if_fail (G_OBJECT_TYPE (old->parent) == GPA_TYPE_LIST, FALSE);
 
 	new = gpa_node_lookup (old->parent, value);
 	if (!new) {
